//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     hyper.cxx
// Purpose:    Hyperresolution (for deriving conditional facts)
// Last Change:12.07.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2001-2002 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include <stdlib.h>
#include "out.h"
#include "check.h"
#include "opt.h"
#include "op.h"
#include "cf.h"
#include "rlit.h"
#include "rlits.h"
#include "rule.h"
#include "rules.h"

#include "hyper.h"

//=============================================================================
// Class Variables:
//=============================================================================

//-----------------------------------------------------------------------------
// Rules: The current set of rules.
//-----------------------------------------------------------------------------

rules_t hyper_c::Rules = 0;

//-----------------------------------------------------------------------------
// Last: The last rule in the current set (to append the next rule).
//-----------------------------------------------------------------------------

rules_t hyper_c::Last = 0;

//-----------------------------------------------------------------------------
// Bindings: Variable bindings for current rule.
//-----------------------------------------------------------------------------

cnst_t hyper_c::Bindings[VER_MAX_VARS];

//-----------------------------------------------------------------------------
// Matches: What are current matches for the body literals.
//-----------------------------------------------------------------------------

cf_hd_t hyper_c::Matches[VER_MAX_BODY_LIT];

//-----------------------------------------------------------------------------
// MatchPtr: Stackpointer for Matches (points to next free element).
//-----------------------------------------------------------------------------

cf_hd_t *hyper_c::MatchPtr;

//-----------------------------------------------------------------------------
// RuleApplied: A rule was applied in this iteration round.
//-----------------------------------------------------------------------------

bool_t hyper_c::RuleApplied;

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// store: Add a rule to the current set of rules.
//-----------------------------------------------------------------------------

void hyper_c::store(rule_t rule)
{
	rules_t new_elem;

	// Check Parameter:
	CHECK(rule != 0, "hyper_c::store: rule is NIL");
	CHECK(rule->valid(), "hyper_c::store: rule is invalid");

	// Create new list element:
	new_elem = new rules_c(rule);

	// Link it to the list:
	if(Last == 0) {
		Rules = new_elem;
		Last = new_elem;
	}
	else {
		Last->append(new_elem);
		Last = new_elem;
	}
}

//-----------------------------------------------------------------------------
// print: Print all rules.
//-----------------------------------------------------------------------------

void hyper_c::print_rules(void)
{

	// Print the headline:
	out_c::rules();

	// Print the rules, handle the case of the empty set:
	if(Rules == 0)
		out_c:: no_rules();
	else
		Rules->print();
}

//-----------------------------------------------------------------------------
// compute_implied: Compute hyperresolution fixpoint (implied cond facts).
//-----------------------------------------------------------------------------

void hyper_c::compute_implied(void)
{
	rules_t r;
	rule_t rule;
	int round;

	// Print headline for verbose output:
	if(opt_c::is_set(OPT_HYPERRES))
		out_c::hyperres_verbose();

	// Move the given facts from the new into the diff status:
	cf_c::next_iter();

	// As long as there are changes, apply all rules:
	round = 1;
	while(cf_c::diff_facts() != 0) {
		if(opt_c::is_set(OPT_HYPERRES)) {
			out_c::hyperres_round(round);
			out_c::list_begin();
			RuleApplied = BOOL_FALSE;
		}
		for(r = Rules; r; r = r->get_tail()) {
			rule = r->get_head();
			MatchPtr = Matches;
			apply_rule(rule, rule->get_body(), BOOL_FALSE);
			cf_c::delete_obsolete();
		}
		if(opt_c::is_set(OPT_HYPERRES)) {
			if(!RuleApplied) {
				out_c::list_item();
				out_c::no_match();
			}
			out_c::list_end();
		}
		if(RuleApplied) {
			out_c::hyperres_curr_cf();
			cf_c::print_all();
		}
		cf_c::next_iter();
		round++;
	}
}

//=============================================================================
// Auxillary Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// apply_rule: Compute implied conditional facts for a given rule.
//-----------------------------------------------------------------------------

void hyper_c::apply_rule(rule_t rule, rlits_t body_rest, bool_t diff_used)
{
	cf_t cf;
	rlit_t body_lit;
	rlits_t body_cont;
	glit_t body_fact;
	cf_hd_t hd;

	// If all body literals are matched, generate conditional fact:
	if(body_rest == 0) {
		generate_cf(rule);
		return;
	}

	// Get the body literal to be matched and the list of remaining ones:
	body_lit = body_rest->get_head();
	body_cont = body_rest->get_tail();

	// Optimized version for ground body literals:
	if(body_lit->is_ground()) {
		body_fact = body_lit->get_ground();
	
		// Are matches with old facts acceptable?
		if(diff_used || body_cont != 0) {
			for(cf = cf_c::old_facts(); cf != 0;
					cf = cf->get_next()) {
				for(hd = cf->get_head(); hd != 0;
						hd = hd->get_next_head_elem()){
					if(body_fact == hd->get_fact()) {
						push_match(hd);
						apply_rule(rule, body_cont,
							diff_used);
						pop_match();
					}
				}
			}
		}

		// Try matches with facts generated in the last iteration:	
		for(cf = cf_c::diff_facts(); cf != 0;
				cf = cf->get_next()) {
			for(hd = cf->get_head(); hd != 0;
					hd = hd->get_next_head_elem()){
				if(body_fact == hd->get_fact()) {
					push_match(hd);
					apply_rule(rule, body_cont, BOOL_TRUE);
					pop_match();
				}
			}
		}

	}

	// Standard version (body literal is not ground):
	else {
	
		// Are matches with old facts acceptable?
		if(diff_used || body_cont != 0) {
			for(cf = cf_c::old_facts(); cf != 0;
					cf = cf->get_next()) {
				for(hd = cf->get_head(); hd != 0;
						hd = hd->get_next_head_elem()){
					if(match(body_lit, hd->get_fact())) {
						push_match(hd);
						apply_rule(rule, body_cont,
							diff_used);
						pop_match();
					}
				}
			}
		}

		// Try matches with facts generated in the last iteration:	
		for(cf = cf_c::diff_facts(); cf != 0;
				cf = cf->get_next()) {
			for(hd = cf->get_head(); hd != 0;
					hd = hd->get_next_head_elem()){
				if(match(body_lit, hd->get_fact())) {
					push_match(hd);
					apply_rule(rule, body_cont, BOOL_TRUE);
					pop_match();
				}
			}
		}
	}
}

//-----------------------------------------------------------------------------
// match: Try to unify a body literal with a fact:
//-----------------------------------------------------------------------------

bool_t hyper_c::match(rlit_t lit, glit_t fact)
{
	rargs_t lit_args;
	rarg_t  lit_arg;
	garg_t fact_args;

	// Compare the predicate of rule literal and fact:
	if(lit->get_pred() != fact->get_pred())
		return(BOOL_FALSE);

	// Compare the argument lists:
	fact_args = fact->get_args();
	for(lit_args = lit->get_args(); lit_args != 0;
			lit_args = lit_args->get_tail()) {
		lit_arg = lit_args->get_head();
		
		// Check whether argument list of fact is shorter:
		if(fact_args == 0) {
			return(BOOL_FALSE);
		}

		// Compare argument terms, consider different kinds:
		if(lit_arg->is_cnst()) {
			if(lit_arg->get_cnst()->
				equal_to(fact_args->get_cnst()) == BOOL_FALSE)
				return(BOOL_FALSE);
		}
		else if(lit_arg->is_bound()) {
			if(Bindings[lit_arg->get_offset()]->
				equal_to(fact_args->get_cnst()) == BOOL_FALSE)
				return(BOOL_FALSE);
		}
		else if(lit_arg->is_free()){
			Bindings[lit_arg->get_offset()] = fact_args->get_cnst();
		}
		else {
			CHECK(lit_arg->is_anon(),
				"hyper_c::match: Unkown kind of argument");
			// Nothing to do
		}

		// Switch to next argument term of fact:
		fact_args = fact_args->get_next();
	}

	// Check whether argument list of fact is longer:
	if(fact_args != 0)
		return(BOOL_FALSE);

	// Ok, the literal and the fact match:
	return(BOOL_TRUE);
}

//-----------------------------------------------------------------------------
// generate_cf: Matches for body literals were found, now generate derived cf.
//-----------------------------------------------------------------------------

void hyper_c::generate_cf(rule_t rule)
{
	glit_t	derived_head[VER_MAX_HEAD_LIT+1];
	glit_t	*derived_head_ptr;
	glit_t	fact;
	rlits_t	rule_head_elem;
	rlit_t	rule_head_lit;
	cf_hd_t	*match_ptr;
	cf_hd_t match_head_elem;

	gneg_t	derived_cond[VER_MAX_COND_NEG+1];
	gneg_t	*derived_cond_ptr;
	rneg_t	rule_neg;
	gneg_t	ground_neg;
	cf_cn_t	match_cond_elem;

	//-------------------------------
	// Construct Head:
	//-------------------------------

	// Open new set of facts for head (to eliminate duplicate head lits):
	glit_c::dup_ck_begin();

	// Initialize pointer to temporary storage of output head literals:
	derived_head_ptr = derived_head;

	// Construct instances of head literals of the rule:
	for(rule_head_elem = rule->get_head(); rule_head_elem != 0;
			rule_head_elem = rule_head_elem->get_tail()) {
		// Get ground instance of head literal in rule:
		rule_head_lit = rule_head_elem->get_head();
		if(rule_head_lit->is_ground())
			fact = rule_head_lit->get_ground();
		else
			fact = lit_instance(rule_head_lit);

		// Add to head of derived conditional fact:
		if(fact->dup_ck_is_new()) {
			CHECK(derived_head_ptr >= derived_head,
				"hyper_c::generate_cf: head_ptr < head");
			CHECK(derived_head_ptr < derived_head+VER_MAX_HEAD_LIT,
				"hyper_c::generate_cf: head_ptr too large");
			*derived_head_ptr++ = fact;
		}
	}

	// Add disjunctive contexts of matches for body literals:
	for(match_ptr = Matches; match_ptr < MatchPtr; match_ptr++) {
		for(match_head_elem = (*match_ptr)->get_cf()->get_head();
			match_head_elem != 0;
			match_head_elem =
				match_head_elem->get_next_head_elem()) {
			if(match_head_elem != *match_ptr) {
				fact = match_head_elem->get_fact();
				if(fact->dup_ck_is_new()) {
					if(derived_head_ptr ==
						derived_head+VER_MAX_HEAD_LIT){
						out_c::too_many_head_lit
							(VER_MAX_HEAD_LIT);
						exit(4);
					}
					CHECK(derived_head_ptr >= derived_head,
						"hyper_c::generate_cf: [1]");
					CHECK(derived_head_ptr <
						derived_head+VER_MAX_HEAD_LIT,
						"hyper_c::generate_cf: [2]");
					*derived_head_ptr++ = fact;
				}
			}
		}
	}

	// Terminate set of head literals:
	*derived_head_ptr = 0;

	// End of duplicate elimination in the derived head:
	glit_c::dup_ck_end();

	//-------------------------------
	// Construct Condition:
	//-------------------------------

	// Open new set of default negations (to eliminate duplicates):
	gneg_c::dup_ck_begin();

	// Initialize pointer to temporary storage of output negations:
	derived_cond_ptr = derived_cond;

	// Construct instances of default negations in the rule:
	for(rule_neg = rule->get_cond(); rule_neg != 0;
			rule_neg = rule_neg->get_next()) {

		// Get ground instance of default negation in rule:
		if(rule_neg->is_ground())
			ground_neg = rule_neg->get_ground();
		else
			ground_neg = neg_instance(rule_neg);

		// Add to condition of derived conditional fact:
		if(ground_neg->dup_ck_is_new()) {
			CHECK(derived_cond_ptr >= derived_cond,
				"hyper_c::generate_cf: cond_ptr < cond");
			CHECK(derived_cond_ptr < derived_cond+VER_MAX_COND_NEG,
				"hyper_c::generate_cf: cond_ptr too large");
			*derived_cond_ptr++ = ground_neg;
		}
	}

	// Add conditions of matches for body literals:
	for(match_ptr = Matches; match_ptr < MatchPtr; match_ptr++) {
		for(match_cond_elem = (*match_ptr)->get_cf()->get_cond();
			match_cond_elem != 0;
			match_cond_elem =
				match_cond_elem->get_next_cond_elem()) {
			ground_neg = match_cond_elem->get_neg();
			if(ground_neg->dup_ck_is_new()) {
				if(derived_cond_ptr ==
					derived_cond+VER_MAX_COND_NEG){
					out_c::too_many_cond_neg
							(VER_MAX_COND_NEG);
					exit(4);
				}
				CHECK(derived_cond_ptr >= derived_cond,
					"hyper_c::generate_cf: [3]");
				CHECK(derived_cond_ptr <
					derived_cond+VER_MAX_COND_NEG,
					"hyper_c::generate_cf: [4]");
				*derived_cond_ptr++ = ground_neg;
			}
		}
	}

	// Terninate set of default negations:
	*derived_cond_ptr = 0;

	// End of duplicate elimination in the derived condition:
	gneg_c::dup_ck_end();

	//-------------------------------
	// Store derived conditional fact:
	//-------------------------------

	cf_c::store(derived_head, derived_head_ptr - derived_head,
			derived_cond, derived_cond_ptr - derived_cond);

	//-------------------------------
	// Verbose output:
	//-------------------------------

	if(opt_c::is_set(OPT_HYPERRES)) {
		// A rule was successfully applied in this round:
		RuleApplied = BOOL_TRUE;

		// Print derived conditional fact:
		out_c::list_item();
		derived_head_ptr = derived_head;
		if(*derived_head_ptr == 0)
			out_c::print_str("FALSE");
		while(*derived_head_ptr) {
			(*derived_head_ptr)->print();
			derived_head_ptr++;
			if(*derived_head_ptr)
				out_c::print_str(op_c::op_or()->name());
		}
		derived_cond_ptr = derived_cond;
		if(*derived_cond_ptr != 0)
			out_c::print_str(op_c::op_if()->name());
		while(*derived_cond_ptr) {
			(*derived_cond_ptr)->print();
			derived_cond_ptr++;
			if(*derived_cond_ptr)
				out_c::print_str(op_c::op_and()->name());
		}

		// Print applied rule:
		out_c::print_str(".");
		out_c::print_str(" [");
		rule->print();
		out_c::print_str("]");

		// Print conditional facts matched with body literals:
		for(match_ptr = Matches; match_ptr < MatchPtr; match_ptr++) {
			out_c::print_str(" [");
			(*match_ptr)->get_cf()->print();
			out_c::print_str("]");

		}
	}
}

//-----------------------------------------------------------------------------
// lit_instance: Compute instance of atomic formula (e.g. head literal).
//-----------------------------------------------------------------------------

glit_t hyper_c::lit_instance(rlit_t lit)
{
	cnst_t	fact_args[VER_MAX_ARGS+1];
	cnst_t	*fact_arg_ptr;
	rargs_t	lit_args;
	rarg_t	lit_arg;

	// Get ground argument terms, store in fact_args:
	fact_arg_ptr = fact_args;
	for(lit_args = lit->get_args(); lit_args;
			lit_args = lit_args->get_tail()) {
		lit_arg = lit_args->get_head();
		if(lit_arg->is_cnst()) {
			CHECK(fact_arg_ptr >= fact_args,
				"hyper_c::lit_instance: fact_args too small");
			CHECK(fact_arg_ptr < fact_args+VER_MAX_ARGS,
				"hyper_c::lit_instance: fact_args too large");
			*fact_arg_ptr++ = lit_arg->get_cnst();
		}
		else {
			CHECK(lit_arg->is_bound(),
				"hyper_c::lit_instance: instance not ground");
			CHECK(lit_arg->get_offset() >= 0,
				"hyper_c::lit_instance: offset negative");
			CHECK(lit_arg->get_offset() < VER_MAX_VARS,
				"hyper_c::lit_instance: offset too large");
			CHECK(fact_arg_ptr >= fact_args,
				"hyper_c::lit_instance: fact_args too small2");
			CHECK(fact_arg_ptr < fact_args+VER_MAX_ARGS,
				"hyper_c::lit_instance: fact_args too large2");
			CHECK(Bindings[lit_arg->get_offset()]->valid(),
				"hyper_c::lit_instance: invalid binding");
			*fact_arg_ptr++ = Bindings[lit_arg->get_offset()];
		}
	}
	*fact_arg_ptr = 0;

	// Create/retrieve fact:
	return(glit_c::store(lit->get_pred(), fact_args));
}

//-----------------------------------------------------------------------------
// neg_instance: Compute instance of default negation.
//-----------------------------------------------------------------------------

gneg_t hyper_c::neg_instance(rneg_t neg)
{
	rlits_t	given_lits;
	glit_t	ground_lits[VER_MAX_NEG_SIZE+1];
	glit_t	ground_lit;
	int	neg_size;

	// Open set of ground literals (for duplicate elimination):
	glit_c::dup_ck_begin();

	// Loop over literals in given default negation:
	neg_size = 0;
	for(given_lits = neg->get_lits(); given_lits;
			given_lits = given_lits->get_tail()) {
		ground_lit = lit_instance(given_lits->get_head());
		if(ground_lit->dup_ck_is_new()) {
			CHECK(neg_size >= 0,
				"hyper_c::neg_instance: neg_size is invalid");
			CHECK(neg_size < VER_MAX_NEG_SIZE,
				"hyper_c::neg_instance: too many literals");
			ground_lits[neg_size++] = ground_lit;
		}
	}
	ground_lits[neg_size] = 0;

	// End of duplicate elimination:
	glit_c::dup_ck_end();

	// Create or retrieve ground default negation:
	return(gneg_c::store(ground_lits));
}
